import typing as t

from .config import EmailURL
from .exceptions import MailerIsRegisteredError, NotRegisteredMailer
from .message import EmailMessage
from .transports import Transport, create_from_url


class Mailer:
    """A facade for sending mails."""

    def __init__(self, transport: Transport) -> None:
        self.transport = transport

    async def send(self, message: EmailMessage) -> t.Any:
        return await self.transport.send(message)


_mailers: t.Dict[str, Mailer] = {}


def add_mailer(
    url: t.Union[str, EmailURL] = None,
    *,
    mailer: Mailer = None,
    transport: Transport = None,
    name: str = 'default',
) -> None:
    """Add a new mailer. Later you can obtain this mailer using it's name.
    If the name is taken an exception will be raised."""

    if url:
        mailer = Mailer(create_from_url(url))
    elif mailer:
        pass
    elif transport:
        mailer = Mailer(transport)
    else:
        raise ValueError('Either url, or mailer, or transport arguments required.')

    if name in _mailers:
        raise MailerIsRegisteredError('Mailer with name "%s" is already registered.' % name)

    _mailers[name] = mailer


def remove_mailer(name: str) -> None:
    """Remove previously registered mailer."""
    _mailers.pop(name, None)


def clear_mailers() -> None:
    """Remove all mailers."""
    _mailers.clear()


def get_mailer(name: str = 'default') -> Mailer:
    """Get a registered mailer.
    If mailer cannot be found an NotRegisteredMailer will be raised.
    You can register mailer with `add_mailer` function."""

    try:
        return _mailers[name]
    except KeyError:
        raise NotRegisteredMailer('Mailer with name "%s" is not registered.' % name) from None



from __future__ import annotations

import abc
import aiofiles
import aiosmtplib
import datetime
import os
import ssl
import sys
import typing as t
from typing import Any, List, Optional, Union

from .config import EmailURL
from .exceptions import NotRegisteredTransportError
from .message import EmailMessage


class Transport(t.Protocol):  # pragma: nocover
    async def send(self, message: EmailMessage) -> None:
        ...

    @classmethod
    def from_url(cls, url: t.Union[str, EmailURL]) -> t.Optional[Transport]:
        ...


class BaseTransport(abc.ABC):  # pragma: nocover
    @abc.abstractmethod
    async def send(self, message: EmailMessage) -> None:
        raise NotImplementedError()

    @classmethod
    def from_url(cls, url: t.Union[str, EmailURL]) -> t.Optional[Transport]:
        return None


class FileTransport(BaseTransport):
    def __init__(self, directory: str):
        if directory is None or directory == "":
            raise ValueError('Argument "path" of FileTransport cannot be None.')

        self._directory = directory

    async def send(self, message: EmailMessage) -> None:
        file_name = "message_%s.eml" % datetime.datetime.today().isoformat()
        output_file = os.path.join(self._directory, file_name)
        async with aiofiles.open(output_file, "wb") as stream:
            await stream.write(message.as_string().encode("utf8"))

    @classmethod
    def from_url(cls, url: Union[str, EmailURL]) -> FileTransport:
        url = EmailURL(url)
        return cls(url.path)


class NullTransport(BaseTransport):
    async def send(self, message: EmailMessage) -> None:
        pass

    @classmethod
    def from_url(cls, *args: Any) -> NullTransport:
        return cls()


class InMemoryTransport(BaseTransport):
    @property
    def mailbox(self) -> List[EmailMessage]:
        return self._storage

    def __init__(self, storage: List[EmailMessage]):
        self._storage = storage

    async def send(self, message: EmailMessage) -> None:
        self._storage.append(message)

    @classmethod
    def from_url(cls, *args: Any) -> InMemoryTransport:
        mailbox: List[EmailMessage] = []
        return cls(mailbox)


class StreamTransport(BaseTransport):
    def __init__(self, output: Any):
        self._output = output

    async def send(self, message: EmailMessage) -> None:
        self._output.write(str(message))


class ConsoleTransport(StreamTransport):
    def __init__(self) -> None:
        super().__init__(sys.stdout)


class SMTPTransport(BaseTransport):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 25,
        user: Optional[str] = None,
        password: Optional[str] = None,
        use_tls: Optional[bool] = None,
        timeout: int = 10,
        key_file: Optional[str] = None,
        cert_file: Optional[str] = None,
    ):
        self._host = host
        self._user = user
        self._port = port
        self._password = password
        self._use_tls = use_tls
        self._timeout = timeout
        self._key_file = key_file
        self._cert_file = cert_file

    async def send(self, message: EmailMessage) -> None:
        context = ssl.create_default_context()
        client = aiosmtplib.SMTP(
            hostname=self._host,
            port=self._port,
            use_tls=self._use_tls,
            username=self._user,
            password=self._password,
            timeout=self._timeout,
            tls_context=context,
            client_key=self._key_file,
            client_cert=self._cert_file,
        )
        async with client:
            await client.send_message(message.build_message())

    @classmethod
    def from_url(cls, url: Union[str, EmailURL]) -> SMTPTransport:
        url = EmailURL(url)

        def _cast_to_bool(value: str) -> bool:
            return value.lower() in ["yes", "1", "on", "true"]

        timeout = url.options.get("timeout", None)
        if timeout:
            timeout = int(timeout)

        use_tls = _cast_to_bool(url.options.get("use_tls", ""))
        key_file = url.options.get("key_file", None)
        cert_file = url.options.get("cert_file", None)

        return cls(
            url.hostname or "localhost",
            url.port or 25,
            url.username,
            url.password,
            use_tls=use_tls,
            timeout=timeout,
            key_file=key_file,
            cert_file=cert_file,
        )


class GMailTransport(SMTPTransport):
    def __init__(self, user: Optional[str], password: Optional[str], timeout: int = 10):
        super().__init__(
            host="smtp.gmail.com",
            port=465,
            user=user,
            password=password,
            use_tls=True,
            timeout=timeout,
        )

    @classmethod
    def from_url(cls, url: Union[str, EmailURL]) -> GMailTransport:
        url = EmailURL(str(url) + "@default")
        return cls(url.username, url.password)


class MailgunTransport(SMTPTransport):
    def __init__(self, user: Optional[str], password: Optional[str], timeout: int = 10):
        super().__init__(
            host="smtp.mailgun.org",
            port=465,
            user=user,
            password=password,
            use_tls=True,
            timeout=timeout,
        )

    @classmethod
    def from_url(cls, url: Union[str, EmailURL]) -> MailgunTransport:
        url = EmailURL(str(url) + "@default")
        return cls(url.username, url.password)


_protocol_handlers: t.Dict[str, t.Type[Transport]] = {
    'smtp': SMTPTransport,
    'file': FileTransport,
    'null': NullTransport,
    'memory': InMemoryTransport,
    'console': ConsoleTransport,
    'gmail': GMailTransport,
    'mailgun': MailgunTransport,
}


def create_from_url(url: t.Union[str, EmailURL]) -> Transport:
    """Create a transport instance from URL configuration."""

    url = EmailURL(url)
    if url.transport not in _protocol_handlers:
        raise NotRegisteredTransportError(f'No transport found with scheme "{url.transport}".')

    klass = _protocol_handlers[url.transport]
    instance = klass.from_url(url)
    if instance is None:
        instance = klass()
    return instance


def add_protocol_handler(protocol: str, transport: t.Type[Transport]) -> None:
    """Register a new protocol handler.

    Example:
        import mailers

        class MyTransport:
            async def send(self, email_message: EmailMessage) -> None: ...

        mailers.add_transport('myproto', MyTransport)

        # then you can use it like this:
        mailers.add_mailer('myproto://')
    """
    _protocol_handlers[protocol] = transport



from .mailer import get_mailer
from .message import EmailMessage


async def send(message: EmailMessage, mailer: str = 'default') -> None:
    """Send a message."""
    await get_mailer(mailer).send(message)



from __future__ import annotations

import abc
import datetime
import email
import email.encoders
import typing as t
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from .exceptions import BadHeaderError


class BaseAttachment(abc.ABC):
    def __init__(
        self,
        mime_type: str = "application/octet-stream",
        charset: str = None,
        disposition: str = "attachment",
        content_id: str = None,
        headers: t.Dict[str, str] = None,
    ):
        self.mime_type = mime_type
        self.charset = charset
        self.disposition = disposition
        self.content_id = content_id
        self.headers = headers or {}

        if content_id:
            self.headers["Content-ID"] = f"<{content_id}>"
            self.headers["X-Attachment-ID"] = content_id

    @abc.abstractmethod
    def build(self) -> MIMEBase:  # pragma: nocover
        raise NotImplementedError()

    def as_string(self) -> str:
        return self.build().as_string()

    def __str__(self) -> str:
        return self.as_string()


class Attachment(BaseAttachment):
    def __init__(
        self,
        file_name: str,
        contents: t.Union[str, bytes],
        mime_type: str = "application/octet-stream",
        disposition: str = "attachment",
        charset: str = None,
        content_id: str = None,
        headers: t.Dict[str, str] = None,
    ):
        self.file_name = file_name
        self.contents = contents
        super().__init__(
            mime_type=mime_type,
            charset=charset,
            disposition=disposition,
            content_id=content_id,
            headers=headers,
        )

    def read(self) -> t.Union[str, bytes]:
        return self.contents

    def build(self) -> MIMEBase:
        main_type, subtype = self.mime_type.split("/")
        part = MIMEBase(main_type, subtype)
        part.add_header("Content-Disposition", self.disposition, filename=self.file_name)
        part.set_payload(self.read(), self.charset)
        for header_name, header_value in self.headers.items():
            part.add_header(header_name, header_value)
        return part


def _ensure_list(value: t.Union[str, t.Iterable[str], None]) -> t.List[str]:
    if value is None:
        return []

    if isinstance(value, str):
        return [value]
    return list(value)


def _create_address(address: str, name: str = None) -> str:
    if name:
        return f"{name} <{address}>"
    return address


def _forbid_new_lines(value: str) -> str:
    if value is not None and ("\n" in value or "\r" in value):
        raise BadHeaderError(f'Header value "{value}" contains new line characters.')
    return value


class EmailMessage:
    def __init__(
        self,
        *,
        to: t.Union[str, t.Iterable[str]] = None,
        subject: str = None,
        text_body: t.Optional[str] = None,
        from_address: t.Optional[str],
        cc: t.Union[str, t.Iterable[str]] = None,
        bcc: t.Union[str, t.Iterable[str]] = None,
        reply_to: t.Union[str, t.Iterable[str]] = None,
        html_body: t.Optional[str] = None,
        attachments: t.Iterable[Attachment] = None,
        headers: t.Dict[str, str] = None,
        date: t.Optional[datetime.datetime] = None,
        boundary: t.Optional[str] = None,
        charset: t.Optional[str] = None,
        parts: t.Iterable[MIMEBase] = None,
        encoding: str = "quoted-printable",
    ):
        self.to: t.List[str] = _ensure_list(to)
        self.cc = _ensure_list(cc)
        self.bcc = _ensure_list(bcc)
        self.reply_to = _ensure_list(reply_to)

        self.subject = subject
        self.from_address = from_address
        self.text_body = text_body
        self.html_body = html_body
        self.attachments = list(attachments or [])
        self.boundary = boundary
        self.charset = charset
        self.headers = headers or {}
        self.parts = list(parts or [])
        self.encoding = encoding
        self.date = date or datetime.datetime.today()
        self.headers['Date'] = self.date.isoformat()

    def add_to(self, address: str, name: t.Optional[str] = None) -> EmailMessage:
        self.to.append(_create_address(address, name))
        return self

    def add_cc(self, address: str, name: t.Optional[str] = None) -> EmailMessage:
        self.cc.append(_create_address(address, name))
        return self

    def add_bcc(self, address: str, name: t.Optional[str] = None) -> EmailMessage:
        self.bcc.append(_create_address(address, name))
        return self

    def add_reply_to(self, address: str, name: t.Optional[str] = None) -> EmailMessage:
        self.reply_to.append(_create_address(address, name))
        return self

    def add_part(self, part: MIMEBase) -> EmailMessage:
        self.parts.append(part)
        return self

    def add_attachment(self, attachment: Attachment) -> EmailMessage:
        self.attachments.append(attachment)
        return self

    def attach(
        self,
        contents: t.AnyStr,
        file_name: str,
        mime_type: str = "application/octet-stream",
        charset: str = None,
        content_id: str = None,
        headers: t.Dict[str, str] = None,
        disposition: str = "attachment",
    ) -> EmailMessage:
        self.add_attachment(
            Attachment(
                file_name=file_name,
                contents=contents,
                mime_type=mime_type,
                charset=charset,
                content_id=content_id,
                headers=headers,
                disposition=disposition,
            )
        )
        return self

    def build_message(self) -> MIMEMultipart:
        envelope = MIMEMultipart(boundary=self.boundary)

        envelope.preamble = "This is a multi-part message in MIME format."

        if self.subject is not None:
            envelope.add_header("Subject", self.subject)

        # RFC 822
        if self.from_address is None:
            raise BadHeaderError('"from_address" attribute was not set.')

        # RFC 822
        if not len(self.to) and not len(self.bcc):
            raise BadHeaderError('Neither "to" or "bcc" attribute was not set.')

        envelope.add_header("From", _forbid_new_lines(self.from_address))
        envelope.add_header("To", _forbid_new_lines(", ".join(self.to)))
        envelope.add_header("Content-Transfer-Encoding", self.encoding)

        if len(self.cc):
            envelope.add_header("Cc", ", ".join(self.cc))

        if len(self.bcc):
            envelope.add_header("Bcc", ", ".join(self.bcc))

        if len(self.reply_to):
            envelope.add_header("Reply-to", ", ".join(self.reply_to))

        for name, value in self.headers.items():
            envelope.add_header(name, value)

        # create content parts
        main_message = MIMEMultipart("alternative")
        envelope.attach(main_message)

        if self.text_body:
            text_part = MIMEText(self.text_body, "plain", self.charset)
            main_message.attach(text_part)

        if self.html_body:
            html_part = MIMEText(self.html_body, "html", self.charset)
            main_message.attach(html_part)

        # add custom parts and attachments at the end to prevent their content
        # be used as preview message in GMail.
        for part in self.parts:
            envelope.attach(part)

        for attachment in self.attachments:
            part = attachment.build()
            email.encoders.encode_base64(part)
            envelope.attach(part)

        return envelope

    def as_string(self) -> str:
        return self.build_message().as_string()

    def __str__(self) -> str:
        return self.as_string()



class MailersError(Exception):
    """Base error class."""


class NotRegisteredTransportError(MailersError):
    """Raised when transport is not registered but retrieved."""


class BadHeaderError(MailersError):
    """Raised if the header contains errors."""


class MailerIsRegisteredError(MailersError):
    """Raised if the mailer's name is already taken."""


class NotRegisteredMailer(MailersError):
    """Raised when a mailer is not registered."""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mailers']

package_data = \
{'': ['*']}

install_requires = \
['aiosmtplib>=1.1.6,<2.0.0']

setup_kwargs = {
    'name': 'mailers',
    'version': '1.0.0',
    'description': 'Email delivery for asyncio.',
    'long_description': '# Mailers for asyncio\n\n![PyPI](https://img.shields.io/pypi/v/mailers)\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/alex-oleshkevich/mailers/Lint)\n![GitHub](https://img.shields.io/github/license/alex-oleshkevich/mailers)\n![Libraries.io dependency status for latest release](https://img.shields.io/librariesio/release/pypi/mailers)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/mailers)\n![GitHub Release Date](https://img.shields.io/github/release-date/alex-oleshkevich/mailers)\n![Lines of code](https://img.shields.io/tokei/lines/github/alex-oleshkevich/mailers)\n\n\n## Installation\n\n```bash\npip install mailers\n```\n\n## Usage\n\nThe package uses two main concepts: mailers and transports. The mailer is a class which abstracts you from the\nunderlying transport and the transport does the actual message delivery.\n\n```python\nfrom mailers import add_mailer, EmailMessage, send\n\nadd_mailer(\'smtp://user:password@localhost:25?timeout=2\', name=\'default\')\n\nmessage = EmailMessage(\n    to=\'user@localhost\', from_address=\'from@localhost\',\n    subject=\'Hello\', text_body=\'World!\'\n)\nawait send(\'user@localhost\', message)\n```\n\nOr if you prefer more control on what is going one, take this more verbose path:\n\n```python\nfrom mailers import Mailer, SMTPTransport, EmailMessage\n\nmailer = Mailer(SMTPTransport(\'localhost\', 25))\n\nmessage = EmailMessage(\n    to=\'user@localhost\', from_address=\'from@localhost\',\n    subject=\'Hello\', text_body=\'World!\'\n)\nawait mailer.send(message)\n```\n\n## Compose messages\n\nThe arguments and methods of `EmailMessage ` class are self-explanatory so here is some basic example:\n\n```python\nfrom mailers import EmailMessage, Attachment\n\nmessage = EmailMessage(\n    to=\'user@localhost\',\n    from_address=\'from@example.tld\',\n    cc=\'cc@example.com\',\n    bcc=[\'bcc@example.com\'],\n    text_body=\'Hello world!\',\n    html_body=\'<b>Hello world!</b>\',\n    attachments=[\n        Attachment(\'CONTENTS\', \'file.txt\', \'text/plain\'),\n    ]\n)\n\n# attachments can be added on demand:\n\nwith open(\'file.txt\', \'r\') as f:\n    message.attach(f.read(), f.name, \'text/plain\')\n\n    # alternatively\n    message.add_attachment(\n        Attachment(f.read(), f.name, \'text/plain\')\n    )\n```\n\n`cc`, `bcc`, `to`, `reply_to` can be either strings or lists of strings.\n\n### A note about attachments\n\nAccessing files is a blocking operation. You may want to use `aiofiles` or alternate library which reads files in\nnon-blocking mode.\n\nThis package does not implement direct access to files at moment. This is something to do at later stage.\n\n## Transports\n\n### Preinstalled transports\n\nAll transport classes can be found in `mailers.transports` module.\n\n| Class             | Example URL                                       | Description                                                       |\n|-------------------|---------------------------------------------------|-------------------------------------------------------------------|\n| SMTPTransport     | smtp://user:pass@hostname:port?timeout=&use_tls=1 | Sends mails using SMTP protocol.                                  |\n| InMemoryTransport | not available                                     | Stores sent messages in the local variable. See an example below. |\n| FileTransport     | file:///path/to/directory                         | Writes sent messages into directory.                              |\n| NullTransport     | null://                                           | Does not perform any sending.                                     |\n| StreamTransport   | not available                                     | Writes message to an open stream. See an example below.           |\n| ConsoleTransport  | console://                                        | Prints messages into stdout.                                      |\n| GMailTransport    | gmail://username:password                         | Sends via GMail.                                                  |\n| MailgunTransport  | mailgun://username:password                       | Sends via Mailgun.                                                |\n\n### Special notes\n\n#### InMemoryTransport\n\n`InMemoryTransport` takes a list and writes outgoing mail into it. Read this list to inspect the outbox.\n\n```python\nfrom mailers import InMemoryTransport, EmailMessage\n\nmessage = EmailMessage(from_address=\'noreply@localhost\')\nmailbox = []\ntransport = InMemoryTransport(mailbox)\nawait transport.send(message)\n\nassert message in mailbox\n```\n\n#### StreamTransport\n\nWrites messages into the open stream.\n\n```python\nfrom mailers import StreamTransport, EmailMessage\nfrom io import StringIO\n\nmessage = EmailMessage(from_address=\'noreply@localhost\')\n\ntransport = StreamTransport(output=StringIO())\nawait transport.send(message)\n```\n\n`output` is any IO compatible object.\n\n### Custom transports.\n\nEach transport must implement `async def send(self, message: EmailMessage) -> None` method. Preferably, inherit\nfrom `BaseTransport` class:\n\n```python\nfrom mailers import BaseTransport, Mailer, EmailMessage\n\n\nclass PrintTransport(BaseTransport):\n    async def send(self, message: EmailMessage) -> None\n        print(str(message))\n\n\nmailer = Mailer(transport=PrintTransport())\n```\n\nIn order to make your transport to accept `EmailURL` instances, your transport class has to implement `from_url`\nclass method:\n\n```python\nfrom mailers import BaseTransport, EmailURL\n\n\nclass PrintTransport(BaseTransport):\n    @classmethod\n    def from_url(cls, url: EmailURL) -> "PrintTransport":\n        return cls()\n```\n\n### Add custom transport protocols.\n\nOnce you build a custom transport you can add it\'s URL to enable URL-based configurations.\n\n```python\nfrom mailers import add_protocol_handler, Mailer, BaseTransport, EmailURL, create_from_url\n\nclass PrintTransport(BaseTransport):\n    @classmethod\n    def from_url(cls, url: EmailURL) -> "PrintTransport":\n        return cls()\n\nadd_protocol_handler(\'myprotocol\', PrintTransport)\n\nmailer = Mailer(transport=create_from_url(\'myprotocol://\'))\n```\n\nNote that the transport must to implement `from_url` method to accept URL parameters. Otherwise it will be constructed\nwithout any arguments passed to the `__init__` method.\n',
    'author': 'alex.oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alex-oleshkevich/mailers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .config import EmailURL
from .delivery import send
from .exceptions import MailersError
from .mailer import Mailer, add_mailer, get_mailer
from .message import Attachment, EmailMessage
from .transports import (
    BaseTransport,
    FileTransport,
    GMailTransport,
    InMemoryTransport,
    MailgunTransport,
    NullTransport,
    SMTPTransport,
    StreamTransport,
    add_protocol_handler,
    create_from_url,
)

__all__ = [
    "InMemoryTransport",
    "SMTPTransport",
    "NullTransport",
    "FileTransport",
    "BaseTransport",
    "StreamTransport",
    "GMailTransport",
    "MailgunTransport",
    "MailersError",
    "EmailMessage",
    "Attachment",
    "EmailURL",
    "Mailer",
    "get_mailer",
    "add_mailer",
    "add_protocol_handler",
    "create_from_url",
    "send",
]
